'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

exports.default = function (_ref) {
  var types = _ref.types;


  /**
   * Used to track variables built during the AST pass. We instantiate these in
   * the `Program` visitor in order to support running the plugin in watch mode
   * or on multiple files.
   *
   * @type Store
   */
  var store = new _Store2.default(_mapping2.default.module.has('fp') ? [_mapping2.default.id, 'lodash/fp'] : [_mapping2.default.id]);

  function buildDeclaratorHandler(key) {
    return function (path) {
      var node = path.node;

      var decNode = node[key];
      if (isIdentifier(decNode, path)) {
        node[key] = store.getValueBy('member', decNode.name);
      }
    };
  }

  function buildExpressionHandler(props) {
    return function (path) {
      var node = path.node;

      _lodash2.default.each(props, function (key) {
        var expNode = node[key];
        if (isIdentifier(expNode, path)) {
          node[key] = store.getValueBy('member', expNode.name);
        }
      });
    };
  }

  function getImportBase(pkgStore) {
    return _lodash2.default.get(pkgStore, 'id') == 'lodash/fp' ? 'fp' : '';
  }

  function isDefaultImport(name) {
    return !!store.getStoreBy('default', name);
  }

  function isIdentifier(node, path) {
    if (types.isIdentifier(node)) {
      var parent = _lodash2.default.get(path.scope.getAllBindings(), [node.name, 'path', 'parent']);
      return isImportDeclaration(parent);
    }
    return false;
  }

  function isImportDeclaration(node) {
    return types.isImportDeclaration(node) && store.has(node.source.value);
  }

  /*--------------------------------------------------------------------------*/

  var visitor = {
    Program: function Program(path, state) {
      var _$assign = _lodash2.default.assign(_mapping2.default, (0, _config2.default)(state.opts));

      var id = _$assign.id;

      if (!id) {
        throw new Error('Cannot find Lodash module');
      }
      if (!store.has(id)) {
        store.set(id);
      }
      // Clear tracked Lodash method imports and variables.
      _importModule2.default.cache.clear();
      store.clear();
    },
    ImportDeclaration: function ImportDeclaration(path) {
      var file = path.hub.file;
      var node = path.node;
      var pkgId = node.source.value;

      var pkgStore = store.get(pkgId);

      if (!pkgStore) {
        return;
      }
      var importBase = getImportBase(pkgStore);
      var defaultMap = pkgStore.get('default');
      var memberMap = pkgStore.get('member');

      // Remove the original import node to be replaced.
      path.remove();

      // Track all the Lodash default and specifier imports in the source.
      _lodash2.default.each(node.specifiers, function (spec) {
        var local = spec.local;

        if (types.isImportSpecifier(spec)) {
          // Handle import specifier, e.g. `import {map} from 'lodash'`.
          var identifier = (0, _importModule2.default)(spec.imported.name, file, importBase, local.name);
          memberMap.set(local.name, identifier);
        } else {
          // Handle default specifier, e.g. `import _ from 'lodash'`.
          defaultMap.set(local.name, true);
        }
      });
    },
    MemberExpression: function MemberExpression(path) {
      var file = path.hub.file;
      var node = path.node;
      var object = node.object;
      var property = node.property;


      if (!isIdentifier(object, path)) {
        return;
      }
      var pkgStore = store.getStoreBy('default', object.name);
      if (pkgStore) {
        if (property.name == 'chain') {
          throw new Error(CHAIN_ERROR);
        }
        // Transform `_.foo` to `_foo`.
        path.replaceWith((0, _importModule2.default)(property.name, file, getImportBase(pkgStore)));
      } else {
        // Allow things like `bind.placeholder = {}`.
        node.object = store.getValueBy('member', object.name);
      }
    },
    CallExpression: function CallExpression(path) {
      var file = path.hub.file;
      var node = path.node;
      var callee = node.callee;


      if (isIdentifier(callee, path)) {
        if (isDefaultImport(callee.name)) {
          // Detect chain sequences with `_(value)`.
          throw new Error(CHAIN_ERROR);
        }
        // Update the import specifier if it's marked for replacement.
        node.callee = store.getValueBy('member', callee.name);
      } else if (types.isMemberExpression(callee)) {
        visitor.MemberExpression(path.get('callee'));
      }
      // Support lodash methods as parameters, e.g. `_.flow(_.map, _.head)`.
      _lodash2.default.each(node.arguments, function (arg, index, args) {
        if (isIdentifier(arg, path)) {
          // Assume default imports are placeholders.
          args[index] = isDefaultImport(arg.name) ? types.memberExpression(node.callee, types.identifier('placeholder')) : store.getValueBy('member', arg.name);
        }
      });
    },
    ExportNamedDeclaration: function ExportNamedDeclaration(path) {
      var node = path.node;
      var file = path.hub.file;

      var pkgId = _lodash2.default.get(node, 'source.value');
      var pkgStore = store.get(pkgId);
      var importBase = getImportBase(pkgStore);

      if (pkgStore) {
        node.source = null;
      }
      _lodash2.default.each(node.specifiers, function (spec) {
        var local = spec.local;

        if (pkgStore) {
          spec.local = (0, _importModule2.default)(local.name, file, importBase);
        } else if (isIdentifier(spec.local, path)) {
          spec.local = store.getValueBy('member', local.name);
        }
      });
    },


    // Various other (less common) ways to use a lodash specifier. This code
    // doesn't apply to uses on a lodash object, only directly imported specifiers.
    'Property': buildDeclaratorHandler('value'),
    'VariableDeclarator': buildDeclaratorHandler('init'),

    // Allow things like `o.a = _.noop`.
    'AssignmentExpression': buildExpressionHandler(['right']),

    // Allow things like `var x = y || _.noop`.
    'LogicalExpression': buildExpressionHandler(['left', 'right']),

    // Allow things like `var x = y ? _.identity : _.noop`.
    'ConditionalExpression': buildExpressionHandler(['test', 'consequent', 'alternate'])
  };

  return { visitor: visitor };
};

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _config = require('./config');

var _config2 = _interopRequireDefault(_config);

var _importModule = require('./importModule');

var _importModule2 = _interopRequireDefault(_importModule);

var _mapping = require('./mapping');

var _mapping2 = _interopRequireDefault(_mapping);

var _Store = require('./Store');

var _Store2 = _interopRequireDefault(_Store);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/** The error message used when chain sequences are detected. */
var CHAIN_ERROR = ['Lodash chain sequences are not supported by babel-plugin-lodash.', 'Consider substituting chain sequences with _.flow composition patterns.', 'See https://medium.com/making-internets/why-using-chain-is-a-mistake-9bc1f80d51ba'].join('\n');

/*----------------------------------------------------------------------------*/

;
module.exports = exports['default'];